/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvparamwidget.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVPARAMWIDGET_H
#define QVPARAMWIDGET_H

#include <QWidget>
#include <QToolButton>

#include <qvcore/qvpropertyholder.h>
#include <qvcore/qvworker.h>

class QLineEdit;
class QSlider;
class QwtSlider;
class QCheckBox;

/// \cond

/// @class QVIntParamWidget qvparamwidget.h qvgui/qvparamwidget.h
/// @brief QVIntParamWidget class.
///
/// This is a widget that presents an edit box and a slider to the user, where
/// he can inspect and modify the integer value of an integer type property,
/// from a QVPropertyHolder object.
/// @ingroup qvgui
class QVIntParamWidget: public QWidget
	{
	Q_OBJECT
	public:
		/// @brief Creates a new QVIntParamWidget object
		/// @param orig_holder pointer to property holder which contains the
		/// parameter that should be inspected.
		/// @param gui_holder pointer to property holder to which the widget
		/// will be associated. It could be simply an object which inherits both
		/// from QVIntParamWidget and QVPropertyHolder, or a bigger widget that
		/// contains the QVIntParamWidget (commonly, a 
		/// @ref QVParamsInspectorWidget).
		/// @param property name of the property that will be inspected.
		/// @param parent parent of the widget.
		QVIntParamWidget(QVPropertyHolder *orig_holder, QVPropertyHolder *gui_holder, const QString property, QWidget *parent = 0);

	signals:
		/// @brief This signal is emitted when the value of the int parameter
		/// has been changed by the user.
		void valueChanged(int);

	private slots:
		void setValue();

	private:
		int value,min,max;
		QVPropertyHolder *orig_holder,*gui_holder;
		QString property;
		QLineEdit *lineedit;
		QSlider *slider;
	};

/// @class QVDoubleParamWidget qvparamwidget.h qvgui/qvparamwidget.h
/// @brief QVDoubleParamWidget class.
///
/// This is a widget that presents an edit box and a slider to the user, where
/// he can inspect and modify the value of a double type property,
/// from a QVPropertyHolder object.
/// @ingroup qvgui
class QVDoubleParamWidget: public QWidget
	{
	Q_OBJECT
	public:
		/// @brief Creates a new QVDoubleParamWidget object
		/// @param orig_holder pointer to property holder which contains the
		/// parameter that should be inspected.
		/// @param gui_holder pointer to property holder to which the widget
		/// will be associated. It could be simply an object which inherits both
		/// from QVDoubleParamWidget and QVPropertyHolder, or a bigger widget
		/// that contains the QVDoubleParamWidget (commonly, a 
		/// @ref QVParamsInspectorWidget).
		/// @param property name of the property that will be inspected.
		/// @param parent parent of the widget.
		QVDoubleParamWidget(QVPropertyHolder *orig_holder, QVPropertyHolder *gui_holder, const QString property, QWidget *parent = 0);

	signals:
		/// @brief This signal is emitted when the value of the int parameter
		/// has been changed by the user.
		void valueChanged(double);

	private slots:
		void setValue();

	private:
		double value,min,max;
		QVPropertyHolder *orig_holder,*gui_holder;
		QString property;
		QLineEdit *lineedit;
		QwtSlider *qwtslider;
	};


/// @class QVBoolParamWidget qvparamwidget.h qvgui/qvparamwidget.h
/// @brief QVBoolParamWidget class.
///
/// This is a widget that presents a check box to the user, where
/// he can inspect and modify the value of a bool type property,
/// from a QVPropertyHolder object.
/// @ingroup qvgui
class QVBoolParamWidget: public QWidget
	{
		Q_OBJECT
	public:	
		/// @brief Creates a new QVBoolParamWidget object
		/// @param orig_holder pointer to property holder which contains the
		/// parameter that should be inspected.
		/// @param gui_holder pointer to property holder to which the widget
		/// will be associated. It could be simply an object which inherits both
		/// from QVBoolParamWidget and QVPropertyHolder, or a bigger widget
		/// that contains the QVBoolParamWidget (commonly, a 
		/// @ref QVParamsInspectorWidget).
		/// @param property name of the property that will be inspected.
		/// @param parent parent of the widget.
		QVBoolParamWidget(QVPropertyHolder *orig_holder, QVPropertyHolder *gui_holder, const QString property, QWidget *parent = 0);
	
	signals:
		/// @brief This signal is emitted when the value of the bool parameter
		/// has been changed by the user.
		void valueChanged(bool);

	private slots:
		void setValue();
	
	private:
		bool value;
		QVPropertyHolder *orig_holder,*gui_holder;
		QString property;
		QCheckBox *checkbox;
	};

/// @class QVWorkerTriggerWidget qvparamwidget.h qvgui/qvparamwidget.h
/// @brief QVBoolParamWidget class.
///
/// This is a widget that presents a button to the user that, when it is
/// clicked, emits a valueChanged(QString), where the QString is the name
/// given to the trigger.
/// @ingroup qvgui
class QVWorkerTriggerWidget: public QWidget
	{
	Q_OBJECT

	public:
		/// @brief Creates a new QVWorkerTriggerWidget object
		/// @param worker pointer to the associated worker.
		/// @param property name of the trigger that will be emitted.
		/// @param parent parent of the widget.
		QVWorkerTriggerWidget(QVWorker *worker, const QString triggername, QWidget *parent = 0);

	signals:
		/// @brief This signal is emitted when the corresponding button has
		///been pressed.
		void valueChanged(QString);

	private slots:
		void setValue();

	private:
		QVWorker *worker;
		QString triggername;
		QToolButton *toolbutton;
	};

/// \endcond
#endif

