/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvmplayercamera.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef MPLAYERCAMERA_H
#define MPLAYERCAMERA_H

#include <QObject>
#include <QProcess>
#include <QString>
#include <QList>
#include <QUrl>
#include <QThread>
#include <QMutex>
#include <QWaitCondition>
#include <QFile>
#include <QTimer>
#include <QQueue>

#include <qvcore/qvcamera.h>
#include <qvutils/qnamedpipe.h>

// Auxiliary QVCheckOKMPlayerCamera Class.
// This is an internal convenience thread used only to unblock the open function
// when it is waiting for mplayer to read from the fifo, in the case when
// mplayer aborted when starting (due to bad URL, or wrong file format, for
// example).

/// @cond 	Doxygen will ignore from here...
class QVCheckOKMplayerCamera: public QThread
{
	Q_OBJECT
	public:
		QVCheckOKMplayerCamera(QFile & fifo_file,int max_time_ms_to_wait_for_open);
		~QVCheckOKMplayerCamera();
		void run();
	private slots:
		void writeErrorInFifo();
	private:
		QFile & _fifo_file;
		int _max_time_ms_to_wait_for_open;
};
/// @endcond

// Auxiliary QVCameraThread Class.
// This is an internal convenience thread used in real time cameras to keep
// updating the frames from mplayer independently of the user grabs.

/// @cond 	Doxygen will ignore from here...
class QVCameraThread: public QThread
	{
	Q_OBJECT
	public:
		QVCameraThread(QObject *object,char *slot);
		~QVCameraThread();
		void run();
	private:
		QObject *_object;
		char *_slot;
	};
/// @endcond

// Auxiliary QVMPlayerIOProcessor class.
// Auxiliary class to communicate through stdin and stdout with mplayer, and
// read data such as the number of rows and cols of the video, the fps, and also
// to send commands such as pausing, ask for current position, and so on.

/// @cond 	Doxygen will ignore from here...
class QVMPlayerIOProcessor: public QObject
	{
	Q_OBJECT
	public:
		QVMPlayerIOProcessor(QProcess *mplayer);
		~QVMPlayerIOProcessor();
		void queueCommandToMPlayer(const QString &, bool ifEmpty = false);
		double fps, speed, time_length, time_pos;
		uInt cols, rows;
	public slots:
		int interpretMPlayerOutput();
	private slots:
		void sendCommandToMPlayer();
	private:
		QQueue<QString> command_queue;
		QMutex command_queue_mutex;
		QProcess *mplayer;
	};
/// @endcond

// Auxiliary QVMPlayerFrameGrabber Class.
/// @cond 	Doxygen will ignore from here...
class QVMPlayerFrameGrabber: public QObject
	{
	Q_OBJECT

	public:
		QVMPlayerFrameGrabber(QVMPlayerIOProcessor *, QFile *, uInt rows, uInt cols, bool isRGB, bool realTime);
		~QVMPlayerFrameGrabber();
		uInt getFrameSize() const { return buf_size; };

		void getQVImageGray(QVImage<uChar> &imgGray) const;
		void getQVImageRGB(QVImage<uChar,3> &imgRGB) const;
		void getQVImageYUV(QVImage<uChar> &imgY, QVImage<uChar> &imgU, QVImage<uChar> &imgV) const;

		void updateFrameBuffer();
		uInt frames_read;
		bool finished;
		QMutex mutex;
		QWaitCondition condition;

	public slots:
		void updateSlot();

	signals:
		void newReadFrameGrabber();
		// This signal is to defer the call to sendCommandToMplayer, and avoid
		// the "QSocketNotifier:: socket notifier cannot be enabled from another
		// thread" warning:
		void sendCommandSignal();

	private:
		QVCameraThread *qvcamerathread;
		QFile *fifoInput;	
		QVMPlayerIOProcessor *mplayerIOProcessor;

		bool debug_in_memcpy, debug_in_updateSlot;
		bool isYUV, realTime;
		uInt buf_size;

		QVImage<uChar> imgY, img_auxY, imgU, img_auxU,  imgV, img_auxV;
		QVImage<uChar,3> imgRGB, img_auxRGB;
		void readToBuffer(uChar *buf_img_aux, uInt buf_size);
		void getNewFrame(QVImage<uChar,3> &img);
		void getNewFrame(QVImage<uChar> &imgY, QVImage<uChar> &imgU, QVImage<uChar> &imgV);
	};
/// @endcond

// This is for doxygen get the detailed description for QVMPlayerCamera correctly.
class QVWorker;
class QVMPlayerIOProcessor;

/*!
@class QVMPlayerCamera qvmplayercamera.h qvmplayercamera/qvmplayercamera.h
@brief Video source class for reading many video file formats, webcams and digital cameras.

This is an implementation of the class @ref QVCamera, based on the program <a href="http://www.mplayerhq.hu/">MPlayer</a>. This program can read videos from a wide variety of media formats, webcams and digital cameras, and the class QVMPlayerCamera takes advantage of that feature. Thus the QVMPlayerCamera is a good, robust and versatile video input reader class.

Internally, when used to read from a video source compatible with MPlayer, the QVMPlayerCamera object will launch a copy of the MPlayer program, and redirect its output to a named pipe, from which will read the frames from the video source.

For a detailed usage of this class in a QVision application, see section @ref ProgrammingModel.

\section QVMPlayerURLFormat Video source identifier URL formats
Each video source is identified by a URL string, even for video files, webcams, remote videos, etc... this class opens a video source, and reads frames from it, until a stop condition arises or the video finishes.

\subsection URLSimplifiedFormat Symplified camera URL format.
Format: just the path to a file or files. Type of formats
 - For video files: 
	\verbatim path-to-file/file-name.{avi, mpg, dv, ...} \endverbatim
 - For devices (webcams, digital cams, tv inputs):
	\verbatim /dev/devicename \endverbatim
 - For image lists:
	\verbatim path-to-directory/filemask.{jpg, png, ...} \endverbatim

\todo Implement and document better this. Mind problems with unspecified camera parameters: fps, brightness, etc... Might be convenient for them to have default values setted in this kind of format.

\subsection ExtendedFormat Extended camera URL formats.
With previous format there can be incompleteness, ambiguous video input, and other problems. More specific URL formats follows:
The format of this string depends of the type of the video source.
Notation: [] means optional, {} means alternative:
	- Local video file compatible with mplayer:
		\verbatim file://[path-to-file/]file-name.{avi, mpg, dv, ...} \endverbatim
		Path can be relative, or absolute. Absoulte paths begin with an extra slash. Examples:
		\verbatim file:///home/vision/penguin.dv \endverbatim
		\verbatim file://../../home/vision/penguin.dv \endverbatim
	- Remote video file (via http, rtsp, ftp):
		\verbatim {http,rtsp,ftp}://domain-name/path-to-file/file.{avi, mpg, dv, ...} \endverbatim
		Examples:
		\verbatim rtsp://ficheros.7rm.es/7RM/7rm06-01-071430.rm \endverbatim
		\verbatim http://perception.inf.um.es/public_data/videos/misc/penguin.dv \endverbatim
		\verbatim ftp://user:passwd@server.dom/data/video.avi \endverbatim

	- DVD | VCD video track (see mplayer man page for correct URL 
		formation for a specific title, track, device, etc.).		
		\verbatim {dvd,vcd}://{video track}\endverbatim
		Examples:
		\verbatim dvd://2 \endverbatim
		\verbatim vcd://1 \endverbatim

	- Sequence of multiple JPEG or PNG files:
		\verbatim mf://path-to/filemask.{jpg, png, ...} \endverbatim
		Examples:
		\verbatim mf:///home/user/images/img*.jpg \endverbatim
		\verbatim mf:///./images/img*.png \endverbatim

	- Video 4 linux, video 4 linux 2 camera, and analog inputs:
		\verbatim {v4l,v4l2,analog}://{device file}?[options] \endverbatim
		Examples:
		\verbatim v4l:///dev/video0 \endverbatim
		\verbatim analog:///dev/video0?height=480&width=640&brightness=-100&fps=25 \endverbatim
		\b Note: Optional specifiers H and W are used by mplayer to open
		the camera in a given source size. Note that this is not the
		same as opening the camera in a default size and then scaling
 		the output image to a given size using the \c r and \c c
		parameters (see overloaded function).
		Check manual page for mplayer, for a complete list of options, at parameter '-tv' option
		specifications.

@ingroup qvmplayercamera
*/

class QVMPlayerCamera : public QVCamera
	{
	Q_OBJECT

	public:
		/// @brief Constructs a QVMPlayerCamera, which is initially in "Closed" status.
		QVMPlayerCamera(QString name = QString());
		/// @brief Destroys a QVMPlayerCamera.
		~QVMPlayerCamera();

		// Camera opening options as flags. Combine using | operator to create
		// an OpenOptions flags (observe plural) object:
		/// @brief Open options for a mplayer camera. Combine them using OR (|).
		enum OpenOption {
			Default = 0x0,		/// Default camera (no realtime, do not discard frames, do not deinterlace, loop mode, produce YUV images)
			RealTime = 0x1,		/// Camera runs in an independent, real time thread.
			Deinterlaced = 0x2,	/// Make mplayer deinterlace the image (odd lines).
			NoLoop = 0x4,		/// If video reaches end, do not loop. Instead, automatically close the camera.
			RGBMEncoder = 0x8	/// Camera will produce interleave RGB images (instead of YUV).
		};

		Q_DECLARE_FLAGS(OpenOptions,OpenOption);

		/// @brief Opens a mplayer source given by its URL.
		/// @param url URL of the mplayer source. See @ref QVMPlayerURLFormat for URL formats.
		/// @param opts Or'ed set of opening options (see QVMPlayerCamera::OpenOption).
		/// @return True if the camera was opened successfully, false otherwise.
		bool openCam(const QString & url,OpenOptions opts = Default);

		/// @brief Opens a mplayer source given by its URL, and scales output to a requested size.
		///
		///	 (Overloaded function, provided by convenience).
		///	 @param url Camera identifying URL string. See @ref QVMPlayerURLFormat for URL formats.
		///	 @param r Number of requested output rows.
		///	 @param c Number of requested output cols.
		///	 @param opts Camera opening options (see QVMPlayerCamera::openCam(const QString&,OpenOptions))
		/// @return True if the camera was opened successfully, false otherwise.
		bool openCam(const QString & urlstring, unsigned int r, unsigned int c, OpenOptions opts = Default);

		/// @brief Opens a mplayer source using the internal properties assigned to it.
		/// (Overloaded function, provided by convenience; see
		/// QVPropertyHolder for details on properties).
		/// Used properties:
		///	 - QString "Video": Camera URL (see QVMPlayerCamera::openCam(const QString&,OpenOptions))
		///	 - int "Rows": Number of requested output rows.
		///	 - int "Cols": Number of requested output cols.
		///	 - bool "RealTime": If the camera should be opened in RealTime mode (see QVMPlayerCamera::openCam(const QString&,OpenOptions)).
		///	 - bool "Deinterlaced": If the camera should be opened in Deinterlaced mode (see QVMPlayerCamera::openCam(const QString&,OpenOptions)).
		///	 - bool "RGBMEncoder": If the camera should be opened in RGBMEncoder mode (see QVMPlayerCamera::openCam(const QString&,OpenOptions)).
		///	 - bool "NoLoop": If the camera should be opened in NoLoop mode (see QVMPlayerCamera::openCam(const QString&,OpenOptions)).
		/// @return True if the camera was opened successfully, false otherwise.
		bool openCam();

		/// @brief Grabs a frame in RGB format.
		///
		/// If the camera is not opened with RGBMencoderMode option, this function will set param QVImage to null image.
		/// @param image QVImage that will store grabbed frame.
		/// @returns TRUE if success, FALSE if fail.
		bool grab(QVImage<uChar,3> &image);

		/// @brief Grabs a gray scale frame.
		///
		/// If the camera is opened with RGBMencoderMode option, this function will set param QVImage to null image.
		/// @param image QVImage that will store grabbed frame.
		/// @returns True if success, false otherwise.
		bool grab(QVImage<uChar,1> &image);

		/// @brief Grabs a frame in YUV format.
		///
		/// This function grabs the image in YUV format, storing each channel Y, U and V, in a separate QVImage of
		/// one channel. YUV format would be 422, thus QVImage corresponding to plane Y will be four times bigger than
		/// those corresponding to planes U and V.
		///
		/// If the camera is opened with RGBMencoderMode option, this function will set param QVImage's to null images.
		/// @param imgY QVImage that will store grabbed frame Y channel.
		/// @param imgU QVImage that will store grabbed frame U channel.
		/// @param imgV QVImage that will store grabbed frame V channel.
		/// @returns True if success, false otherwise.
		bool grab(QVImage<uChar> &imgY, QVImage<uChar> &imgU, QVImage<uChar> &imgV);

		/// @brief Returns the current camera name (without path).
		/// @returns Current camera name.
		const QString getUrlBase() const { return url.path().split("/").last(); }

		/// @brief Returns current camera options.
		/// @returns Current camera options.
		OpenOptions getOptions() const		{ return open_options; };

		/// @brief Returns the current frame buffer size.
		/// @returns Current frame buffer size.
		int getBufferSize() const		{ return mplayerFrameGrabber->getFrameSize(); };

		/// @brief Returns the total number of updated frames.
		/// @returns Current total number of updated frames.
		unsigned int getFramesUpdated() const	{ return mplayerFrameGrabber->frames_read; }

		/// @brief Returns the number of rows of the buffer.
		/// @returns Number of rows.
		unsigned int getRows() const		{ return mplayerIOProcessor->rows; };

		/// @brief Returns the number of cols of the buffer.
		/// @returns Number of cols.
		unsigned int getCols() const		{ return mplayerIOProcessor->cols; };

		/// @brief Returns the camera speed in frames per second.
		/// @returns Camera speed in frames per second.
		double getFPS() const			{ return mplayerIOProcessor->fps; };

		/// @brief Returns the video file length in seconds.
		/// @returns Video length in seconds.
		double getTimeLength() const		{ return mplayerIOProcessor->time_length; };

		/// @brief Returns the current video position in seconds.
		/// @returns Current video position in seconds.
		double getTimePos() const		{ return mplayerIOProcessor->time_pos; };

		/// @brief Returns the current video speed.
		/// @returns Current video speed (as a multiplier of the FPS).
		double getSpeed() const			{ return mplayerIOProcessor->speed; };

		/// @brief Returns number of frames grabbed.
		/// @returns Number of successful frames returned by grab's
		unsigned int getFramesGrabbed() const		{ return frames_grabbed; };

		/// @brief Returns number of frames read.
		/// @returns Number of frames actually read by a camera.
		unsigned int getFramesRead() const		{ return mplayerFrameGrabber->frames_read; };

		/// @brief Tells if the camera is a live camera.
		/// @returns True if camera is a live camera.
		bool isLiveCamera() const		{ return live_camera; };

		/// @brief Grabs a RGB image, from a camera specified with an URI.
		///
		/// This function grabs a frame from a video file or url, in RGB format, without having to create explicitly a
		/// mplayer camera object, open it, skip frames to the desired, and close the camera.
		/// @param uri Camera URL (see QVMPlayerCamera::openCam(const QString&,OpenOptions)).
		/// @param image QVImage that will store grabbed frame.
		/// @param frame frame number, 0 for first frame.
		/// @returns TRUE if success, FALSE if fail.
		static bool getFrame(const QString uri, QVImage<uChar,3> &img, int frame = 0)
			{
			QVMPlayerCamera camera;
			if(camera.openCam(uri, QVMPlayerCamera::RGBMEncoder) == -1)
				return false;
			bool result;
			for (int i=-1; i < frame; i++)
				result = camera.grab(img);
			camera.closeCam();
			return result;
			}

		/// @brief Grabs a gray image, from a camera specified with an URI.
		///
		/// This function grabs a gray scale frame from a video file or url, without having to create explicitly a
		/// mplayer camera object, open it, skip frames to the desired, and close the camera.
		/// @param uri Camera URL (see QVMPlayerCamera::openCam(const QString&,OpenOptions)).
		/// @param image QVImage that will store grabbed frame.
		/// @param frame frame number, 0 for first frame.
		/// @returns TRUE if success, FALSE if fail.
		static bool getFrame(const QString uri, QVImage<uChar,1> &img, int frame = 0)
			{
			QVMPlayerCamera camera;
			if(camera.openCam(uri, QVMPlayerCamera::Default) == -1)
				return false;

			for (int i=-1; i < frame; i++)
				if (!camera.grab(img))
					return false;

			camera.closeCam();

			return true;
			}

		/// @brief Grabs a YUV image, from a camera specified with an URI.
		///
		/// This function grabs a frame from a video file or url, in YUV format, without having to create explicitly a
		/// mplayer camera object, open it, skip frames to the desired, and close the camera.
		/// @param uri Camera URL (see QVMPlayerCamera::openCam(const QString&,OpenOptions)).
		/// @param imgY QVImage that will store grabbed frame Y channel.
		/// @param imgU QVImage that will store grabbed frame U channel.
		/// @param imgV QVImage that will store grabbed frame V channel.
		/// @param frame frame number, 0 for first frame.
		/// @returns TRUE if success, FALSE if fail.
		static bool getFrame(const QString uri, QVImage<uChar> &imgY, QVImage<uChar> &imgU, QVImage<uChar> &imgV,
				int frame = 0)
			{
			QVMPlayerCamera camera;
			if(camera.openCam(uri, QVMPlayerCamera::Default) == -1)
				return false;
			bool result;
			for (int i=-1; i < frame; i++)
				result = camera.grab(imgY, imgU, imgV);
			camera.closeCam();
			return result;
			}

	public slots:
		/// @brief Grabs a frame from camera input.
		///
		/// Grabs an image, and stores it in the corresponding properties of the camera:
		///	"RGB image", "Y channel image", "U channel image", and "V channel image".
		/// @param image QVImage that will store grabbed frame.
		/// @returns TRUE if success, FALSE if fail.
		bool grab();

		/// @brief Pauses the camera.
		void pauseCam();
		/// @brief Unpauses the camera.
		void unpauseCam();
		/// @brief Go to the next frame.
		void nextFrameCam();
		/// @brief Set the camera speed.
		/// @param d Camera speed (as a multiplier of the FPS)
		void setSpeedCam(double d);
		/// @brief Moves the camera file to a given position.
		/// @param type Type of seek: Current=0,Percentage=1,Absolute=2.
		/// @param pos 
		///	- If in \c Current mode: Relative displacement with respect to the current position in seconds.
		///	- If in \c Percentage mode: Percentage of total video length.
		///	- If in \c Absolute mode: Absolute position in seconds.
		void seekCam(QVCamera::TSeekType type, double pos);	
		/// @brief Closes the camera.
		void closeCam();

		bool link(QVWorker *,QString imageName);
		bool link(QVWorker *, const QString imageY, const QString imageU, const QString imageV);

	signals:
		// This signal is to defer the call to sendCommandToMplayer, and avoid
		// the "QSocketNotifier:: socket notifier cannot be enabled from another
		// thread" warning:
		void sendCommandSignal();

	private:
		// this boolean value becomes true if any RGB image is linked to the camera, from a worker.
		bool rgbMode;

		QWaitCondition conditionJustToSleepMiliseconds;

		OpenOptions open_options;
		QVMPlayerFrameGrabber * mplayerFrameGrabber;
		QVMPlayerIOProcessor * mplayerIOProcessor;

		QNamedPipe *namedPipe, *namedPipeAux;	// The named fifo(s);
		QUrl url;				// Camera URL;
		QFile fifo_file,fifo_file_2;		// Named fifo;
		QProcess mplayer, mencoder;             // mplayer and mencoder processes;

		QStringList initMPlayerArgs(unsigned int rows, unsigned int cols);
		bool performGrab();

		bool live_camera;
		int frames_grabbed;
	};

// Camera opening flags | operator:
Q_DECLARE_OPERATORS_FOR_FLAGS(QVMPlayerCamera::OpenOptions)

#endif




